/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "gcc-compat.h"
#include "para.h"

#include "ortp.h"
#include "recv.h"

#include <ortp/ortp.h>
#define CHUNK_SIZE 65536

extern gint msg_to_buf(mblk_t *, char *, gint);

struct private_ortp_recv_data {
	RtpSession *session;
	struct timeval start;
	int count, good, c_bad; /* number of (good) chunks / consecutive bad chunks */
	guint32 timestamp;
};

static int handle_other_packet(struct receiver_node *rn, mblk_t *mp)
{
	char buf[CHUNK_SIZE];
	int ret = -1, code, len = msg_to_buf(mp, buf, CHUNK_SIZE);
	struct private_ortp_recv_data *pord = rn->private_data;

	if (len <= 0)
		return -E_MSG_TO_BUF;
	code = buf[0];
	PARA_INFO_LOG("rcvd ctrl pkg #%i (%d bytes)\n", code, len);
	switch (code) {
	case ORTP_EOF:
		ret = 0;
		goto out;
		break;
	case ORTP_BOF:
		rn->flags &= ~HEADER_WAIT;
		gettimeofday(&pord->start, NULL);
		pord->c_bad = 1; /* force default timeout */
		break;
	case ORTP_EMPTY:
		if (!pord->start.tv_sec)
			gettimeofday(&pord->start, NULL);
		pord->count++;
		pord->c_bad = 0;
		break;
	case ORTP_HEADER:
		if (!rn->flags & HEADER_WAIT)
			break; /* just ignore it */
		rn->flags &= ~HEADER_WAIT;
		memmove(rn->buf, buf + 1, len - 1);
		rn->loaded = len - 1;
		PARA_NOTICE_LOG("have header: %d bytes\n", len - 1);
		pord->c_bad = 1; /* force default timeout */
		gettimeofday(&pord->start, NULL);
		break;
	default:
		PARA_WARNING_LOG("%s", "rcvd unknown control pkg\n");
		goto out;
	}
	ret = 1;
out:
	freemsg(mp);
	return ret;
}

static int ortp_recv_pre_select(struct receiver_node *rn, __unused fd_set *rfds,
		__unused fd_set *wfds, struct timeval *timeout)
{
	struct private_ortp_recv_data *pord = rn->private_data;
	struct timeval ret = {0, 5000}, tmp, now, next_chunk;

	if (!pord->start.tv_sec) {/* not yet started */
		pord->timestamp = 0;
		goto out;
	}
	gettimeofday(&now, NULL);
	if (!pord->good || pord->c_bad) { /* previous read failed, use current ts */
		tv_diff(&now, &pord->start, &tmp);
		pord->timestamp = rtp_session_time_to_ts(pord->session, tv2ms(&tmp));
		goto out; /* and default timeout */
	}
	tv_scale(pord->count, &rn->chunk_tv, &tmp);
	pord->timestamp = rtp_session_time_to_ts(pord->session, tv2ms(&tmp));
	tv_add(&tmp, &pord->start, &next_chunk);
	if (tv_diff(&next_chunk, &now, &ret) < 0) {
		/* overdue, sleep only 1ms */
		ret.tv_sec = 0;
		ret.tv_usec = 1000;
	}
out:
	if (tv_diff(timeout, &ret, NULL) > 0)
		*timeout = ret;
	return -1; /* we did not modify rfds */
}

int ortp_recv_post_select(struct receiver_node *rn, int select_ret,
	__unused fd_set *rfds, __unused fd_set *wfds)
{
	struct private_ortp_recv_data *pord = rn->private_data;
	mblk_t *mp = rtp_session_recvm_with_ts(pord->session, pord->timestamp);
	int ret;

	if (select_ret) /* we're only interested in timeouts */
		return 1;
	if (!mp) {
		pord->c_bad++;
		if (!(pord->c_bad % 100))
			PARA_DEBUG_LOG("did not receive  buffer, ts=%u, count = %d, good = %d, c_bad = %d\n",
				pord->timestamp, pord->count, pord->good, pord->c_bad);
		if (pord->good && pord->c_bad > 500) {
			PARA_ERROR_LOG("%d consecutive bad chunks, giving up\n", pord->c_bad);
			return -E_READ;
		}
		goto out;
	}
	PARA_DEBUG_LOG("have it ts = %d, good = %d\n", pord->timestamp, pord->good);
	if (rtp_session_get_recv_payload_type(pord->session) == PAYLOAD_OTHER)
		return handle_other_packet(rn, mp);
	pord->c_bad = 0;
	pord->count++;
	if (!pord->start.tv_sec)
		gettimeofday(&pord->start, NULL);
	/* okay, we have a chunk of data */
	if (rn->flags & HEADER_WAIT) {
		/*
		 * but we can not use it because we need the audio file header and
		 * don't have it yet. Just ignore the data
		 */
		return 1;
	}
	pord->good++;
	if (rn->loaded >= CHUNK_SIZE)
		return -E_OVERRUN;
	ret = msg_to_buf(mp, rn->buf + rn->loaded, CHUNK_SIZE - rn->loaded);
	if (ret <= 0)
		return -E_MSG_TO_BUF;
	rn->loaded += ret;
out:
	return 1;
}

static void ortp_shutdown(void)
{
//	ortp_global_stats_display();
	ortp_exit();
}

static void ortp_recv_close(struct receiver_node *rn)
{
	struct private_ortp_recv_data *pord = rn->private_data;

	rtp_session_reset(pord->session);
	free(rn->private_data);
	free(rn->buf);
}

static int ortp_recv_open(struct receiver_node *rn)
{
	struct private_ortp_recv_data *pord;

	rn->buf = para_malloc(CHUNK_SIZE);
	rn->private_data = para_calloc(sizeof(struct private_ortp_recv_data));
	pord = rn->private_data;
	pord->session = rtp_session_new(RTP_SESSION_RECVONLY);
	PARA_NOTICE_LOG("receiving from %s:%d\n", rn->host, rn->port);
	rtp_session_set_local_addr(pord->session, rn->host, rn->port);
	rtp_session_set_payload_type(pord->session, PAYLOAD_AUDIO_CONTINUOUS);
	return 1;
}

int ortp_recv_init(struct receiver *r)
{
	r->shutdown = ortp_shutdown;
	r->open = ortp_recv_open;
	r->close = ortp_recv_close;
	r->pre_select = ortp_recv_pre_select;
	r->post_select = ortp_recv_post_select;

	ortp_init();
	ortp_set_debug_file("oRTP", NULL);
	return 1;
}
